# Copyright (C) 2024 Nice Zombies
"""JSONYX module."""
from __future__ import annotations

__all__: list[str] = ["dump", "dumps", "load", "loads"]

from codecs import (
    BOM_UTF8, BOM_UTF16_BE, BOM_UTF16_LE, BOM_UTF32_BE, BOM_UTF32_LE,
)
from os.path import realpath
from typing import TYPE_CHECKING

from typing_extensions import Any, Literal

from jsonyx.decoder import JSONDecoder
from jsonyx.encoder import JSONEncoder
from jsonyx.scanner import JSONSyntaxError

if TYPE_CHECKING:
    from collections.abc import Callable, Container

    from _typeshed import SupportsRead, SupportsWrite


# pylint: disable-next=R0913
def dump(  # noqa: PLR0913
    obj: Any,
    fp: SupportsWrite[str],
    *,
    allow: Container[Literal["nan"] | str] = (),
    ensure_ascii: bool = False,
    indent: int | str | None = None,
    item_separator: str = ", ",
    key_separator: str = ": ",
    sort_keys: bool = False,
) -> None:
    """Serialize object to a JSON formatted file."""
    for chunk in JSONEncoder(
        allow=allow,
        ensure_ascii=ensure_ascii,
        indent=indent,
        item_separator=item_separator,
        key_separator=key_separator,
        sort_keys=sort_keys,
    ).iterencode(obj):
        fp.write(chunk)


# pylint: disable-next=R0913
def dumps(  # noqa: PLR0913
    obj: Any,
    *,
    allow: Container[Literal["nan"] | str] = (),
    ensure_ascii: bool = False,
    indent: int | str | None = None,
    item_separator: str = ", ",
    key_separator: str = ": ",
    sort_keys: bool = False,
) -> str:
    """Serialize object to a JSON formatted string."""
    return JSONEncoder(
        allow=allow,
        ensure_ascii=ensure_ascii,
        indent=indent,
        item_separator=item_separator,
        key_separator=key_separator,
        sort_keys=sort_keys,
    ).encode(obj)


def _decode_bytes(b: bytearray | bytes) -> str:
    encoding: str = "utf-8"
    startswith: Callable[[bytes | tuple[bytes, ...]], bool] = b.startswith
    if startswith((BOM_UTF32_BE, BOM_UTF32_LE)):
        encoding = "utf-32"
    elif startswith((BOM_UTF16_BE, BOM_UTF16_LE)):
        encoding = "utf-16"
    elif startswith(BOM_UTF8):
        encoding = "utf-8-sig"
    elif len(b) >= 4:
        if not b[0]:
            # 00 00 -- -- - utf-32-be
            # 00 XX -- -- - utf-16-be
            encoding = "utf-16-be" if b[1] else "utf-32-be"
        elif not b[1]:
            # XX 00 00 00 - utf-32-le
            # XX 00 00 XX - utf-16-le
            # XX 00 XX -- - utf-16-le
            encoding = "utf-16-le" if b[2] or b[3] else "utf-32-le"
    elif len(b) == 2:
        if not b[0]:
            # 00 XX - utf-16-be
            encoding = "utf-16-be"
        elif not b[1]:
            # XX 00 - utf-16-le
            encoding = "utf-16-le"

    return b.decode(encoding, "surrogatepass")


def load(
    fp: SupportsRead[bytearray | bytes | str],
    *,
    allow: Container[
        Literal["comments", "duplicate_keys", "nan", "trailing_comma"] | str
    ] = (),
) -> Any:
    """Deserialize a JSON file to a Python object."""
    return loads(
        fp.read(), allow=allow, filename=getattr(fp, "name", "<string>"),
    )


def loads(
    s: bytearray | bytes | str,
    *,
    allow: Container[
        Literal["comments", "duplicate_keys", "nan", "trailing_comma"] | str
    ] = (),
    filename: str = "<string>",
) -> Any:
    """Deserialize a JSON document to a Python object."""
    if not filename.startswith("<") and not filename.endswith(">"):
        filename = realpath(filename)

    if not isinstance(s, str):
        s = _decode_bytes(s)
    elif s.startswith("\ufeff"):
        msg: str = "Unexpected UTF-8 BOM"
        raise JSONSyntaxError(msg, filename, s, 0)

    return JSONDecoder(allow=allow).decode(s, filename=filename)



# Copyright (C) 2024 Nice Zombies
"""JSON encoder."""
from __future__ import annotations

__all__: list[str] = ["JSONEncoder"]

import json
from typing import TYPE_CHECKING

from typing_extensions import Literal

if TYPE_CHECKING:
    from collections.abc import Container


# TODO(Nice Zombies): only allow strings as keys
class JSONEncoder(json.JSONEncoder):
    """JSON encoder."""

    # TODO(Nice Zombies): align_items=True
    # pylint: disable-next=R0913
    def __init__(  # noqa: PLR0913
        self,
        *,
        allow: Container[Literal["nan"] | str] = (),
        ensure_ascii: bool = False,
        indent: int | str | None = None,
        item_separator: str = ", ",
        key_separator: str = ": ",
        sort_keys: bool = False,
    ) -> None:
        """Create new JSON encoder."""
        if indent is not None:
            item_separator = item_separator.rstrip()

        super().__init__(
            ensure_ascii=ensure_ascii,
            allow_nan="nan" in allow,
            sort_keys=sort_keys,
            indent=indent,
            separators=(item_separator, key_separator),
        )



# Copyright (C) 2024 Nice Zombies
# TODO(Nice Zombies): replace with _jsonyx
"""JSON accelerator."""
from __future__ import annotations

__all__: list[str] = ["DuplicateKey", "make_scanner", "scanstring"]

# pylint: disable-next=E0611, E0401
from _jsonyx import (  # type: ignore # isort: skip
    DuplicateKey, make_scanner, scanstring,  # type: ignore # noqa: PLC2701
)



# Copyright (C) 2024 Nice Zombies
"""JSON decoder."""
from __future__ import annotations

__all__: list[str] = ["DuplicateKey", "JSONDecoder"]

import re
from re import DOTALL, MULTILINE, VERBOSE, Match, RegexFlag
from typing import TYPE_CHECKING

from jsonyx.scanner import JSONSyntaxError, make_scanner
from typing_extensions import Any, Literal

if TYPE_CHECKING:
    from collections.abc import Callable, Container

FLAGS: RegexFlag = VERBOSE | MULTILINE | DOTALL
UNESCAPE: dict[str, str] = {
    '"': '"',
    "\\": "\\",
    "/": "/",
    "b": "\b",
    "f": "\f",
    "n": "\n",
    "r": "\r",
    "t": "\t",
}

_match_chunk: Callable[[str, int], Match[str] | None] = re.compile(
    r'[^"\\\x00-\x1f]+', FLAGS,
).match


def _unescape_unicode(filename: str, s: str, pos: int) -> int:
    esc: str = s[pos:pos + 4]
    if len(esc) == 4 and esc[1] not in "xX":
        try:
            return int(esc, 16)
        except ValueError:
            pass

    msg: str = "Expecting 4 hex digits"
    raise JSONSyntaxError(msg, filename, s, pos)


try:
    # pylint: disable-next=C0412
    from jsonyx._accelerator import scanstring
except ImportError:
    def scanstring(filename: str, s: str, end: int, /) -> (  # noqa: C901
        tuple[str, int]
    ):
        """Scan JSON string."""
        chunks: list[str] = []
        append_chunk: Callable[[str], None] = chunks.append
        str_idx: int = end - 1
        while True:
            # Match one or more unescaped string characters
            if match := _match_chunk(s, end):
                end = match.end()
                append_chunk(match.group())

            # Terminator is the end of string, a literal control character,
            # or a backslash denoting that an escape sequence follows
            try:
                terminator: str = s[end]
            except IndexError:
                msg: str = "Unterminated string"
                raise JSONSyntaxError(msg, filename, s, str_idx) from None

            if terminator == '"':
                return "".join(chunks), end + 1

            if terminator != "\\":
                if terminator == "\n":
                    msg = "Unterminated string"
                    raise JSONSyntaxError(msg, filename, s, str_idx)

                msg = "Unescaped control character"
                raise JSONSyntaxError(msg, filename, s, end)

            end += 1
            try:
                esc = s[end]
            except IndexError:
                msg = "Invalid backslash escape"
                raise JSONSyntaxError(msg, filename, s, end) from None

            # If not a unicode escape sequence, must be in the lookup table
            if esc != "u":
                try:
                    char = UNESCAPE[esc]
                except KeyError:
                    msg = "Invalid backslash escape"
                    raise JSONSyntaxError(msg, filename, s, end) from None

                end += 1
            else:
                uni: int = _unescape_unicode(filename, s, end + 1)
                end += 5
                if 0xd800 <= uni <= 0xdbff and s[end:end + 2] == r"\u":
                    uni2: int = _unescape_unicode(filename, s, end + 2)
                    if 0xdc00 <= uni2 <= 0xdfff:
                        uni = ((uni - 0xd800) << 10) | (uni2 - 0xdc00)
                        uni += 0x10000
                        end += 6

                char = chr(uni)

            append_chunk(char)


_match_whitespace: Callable[[str, int], Match[str] | None] = re.compile(
    r"[ \t\n\r]+", FLAGS,
).match


def _skip_comments(context: JSONDecoder, filename: str, s: str, end: int) -> (
    int
):
    while True:
        if match := _match_whitespace(s, end):
            end = match.end()

        if (comment_prefix := s[end:end + 2]) == "//":
            if not context.allow_comments:
                msg: str = "Comments aren't allowed"
                raise JSONSyntaxError(msg, filename, s, end)

            if (end := s.find("\n", end + 2)) == -1:
                return len(s)

            end += 1
        elif comment_prefix == "/*":
            if not context.allow_comments:
                msg = "Comments aren't allowed"
                raise JSONSyntaxError(msg, filename, s, end)

            comment_idx: int = end
            if (end := s.find("*/", end + 2)) == -1:
                msg = "Unterminated comment"
                raise JSONSyntaxError(msg, filename, s, comment_idx)

            end += 2
        else:
            return end


try:
    from jsonyx._accelerator import DuplicateKey  # type: ignore
except ImportError:
    class DuplicateKey(str):
        """Duplicate key."""

        __slots__: tuple[()] = ()

        def __hash__(self) -> int:
            """Return hash."""
            return id(self)


# pylint: disable-next=R0913, R0912
def parse_object(  # noqa: C901, PLR0917, PLR0913, PLR0912
    context: JSONDecoder,
    filename: str,
    s: str,
    end: int,
    scan_once: Callable[[str, str, int], tuple[Any, int]],
    memoize: Callable[[str, str], str],
) -> tuple[dict[str, Any], int]:
    """Parse JSON object."""
    end = _skip_comments(context, filename, s, end)
    if (nextchar := s[end:end + 1]) != '"':
        if nextchar != "}":
            msg: str = "Expecting string"
            raise JSONSyntaxError(msg, filename, s, end)

        return {}, end + 1

    result: dict[str, Any] = {}
    while True:
        key_idx: int = end
        key, end = scanstring(filename, s, end + 1)

        if key not in result:
            # Reduce memory consumption
            key = memoize(key, key)
        elif not context.allow_duplicate_keys:
            msg = "Duplicate keys aren't allowed"
            raise JSONSyntaxError(msg, filename, s, key_idx)
        else:
            key = DuplicateKey(key)

        if s[end:end + 1] != ":":
            colon_idx: int = end
            end = _skip_comments(context, filename, s, end)
            if s[end:end + 1] != ":":
                msg = "Expecting ':' delimiter"
                raise JSONSyntaxError(msg, filename, s, colon_idx)

        end = _skip_comments(context, filename, s, end + 1)
        result[key], end = scan_once(filename, s, end)
        if s[end:end + 1] != ",":
            comma_idx: int = end
            end = _skip_comments(context, filename, s, end)
            if (nextchar := s[end:end + 1]) != ",":
                if nextchar != "}":
                    msg = "Expecting ',' delimiter"
                    raise JSONSyntaxError(msg, filename, s, comma_idx)

                return result, end + 1

        comma_idx = end
        end = _skip_comments(context, filename, s, end + 1)
        if (nextchar := s[end:end + 1]) != '"':
            if nextchar != "}":
                msg = "Expecting string"
                raise JSONSyntaxError(msg, filename, s, end)

            if not context.allow_trailing_comma:
                msg = "Trailing comma's aren't allowed"
                raise JSONSyntaxError(msg, filename, s, comma_idx)

            return result, end + 1


def parse_array(
    context: JSONDecoder,
    filename: str,
    s: str,
    end: int,
    scan_once: Callable[[str, str, int], tuple[Any, int]],
) -> tuple[list[Any], int]:
    """Parse JSON array."""
    end = _skip_comments(context, filename, s, end)
    if (nextchar := s[end:end + 1]) == "]":
        return [], end + 1

    values: list[Any] = []
    append_value: Callable[[Any], None] = values.append
    while True:
        value, end = scan_once(filename, s, end)
        append_value(value)
        if s[end:end + 1] != ",":
            comma_idx: int = end
            end = _skip_comments(context, filename, s, end)
            if (nextchar := s[end:end + 1]) != ",":
                if nextchar == "]":
                    return values, end + 1

                msg: str = "Expecting ',' delimiter"
                raise JSONSyntaxError(msg, filename, s, comma_idx)

        comma_idx = end
        end = _skip_comments(context, filename, s, end + 1)
        if s[end:end + 1] == "]":
            if context.allow_trailing_comma:
                return values, end + 1

            msg = "Trailing comma's aren't allowed"
            raise JSONSyntaxError(msg, filename, s, comma_idx)


class JSONDecoder:  # pylint: disable=R0903, R0902
    """JSON decoder."""

    def __init__(
        self,
        *,
        allow: Container[Literal[
            "comments", "duplicate_keys", "nan", "trailing_comma",
        ] | str] = (),
    ) -> None:
        """Create new JSON decoder."""
        self.allow_comments: bool = "comments" in allow
        self.allow_duplicate_keys: bool = "duplicate_keys" in allow
        self.allow_nan: bool = "nan" in allow
        self.allow_trailing_comma: bool = "trailing_comma" in allow
        self.parse_array: Callable[[
            JSONDecoder, str, str, int,
            Callable[[str, str, int], tuple[Any, int]],
        ], tuple[list[Any], int]] = parse_array
        self.parse_object: Callable[[
            JSONDecoder, str, str, int,
            Callable[[str, str, int], tuple[Any, int]],
            Callable[[str, str], str],
        ], tuple[dict[str, Any], int]] = parse_object
        self.parse_string: Callable[
            [str, str, int], tuple[str, int],
        ] = scanstring
        self.scan_once: Callable[
            [str, str, int], tuple[Any, int],
        ] = make_scanner(self)  # type: ignore

    def decode(self, s: str, *, filename: str = "<string>") -> Any:
        """Decode a JSON document."""
        end: int = _skip_comments(self, filename, s, 0)
        try:
            obj, end = self.scan_once(filename, s, end)
        except JSONSyntaxError as exc:
            raise exc.with_traceback(None) from None

        if (end := _skip_comments(self, filename, s, end)) < len(s):
            msg: str = "Extra data"
            raise JSONSyntaxError(msg, filename, s, end)

        return obj



# !/usr/bin/env python
# Copyright (C) 2024 Nice Zombies
"""JSON tool."""
from __future__ import annotations

__all__ = ["JSONNamespace", "register", "run"]

import sys
from argparse import ArgumentParser
from pathlib import Path
from sys import stdin

from jsonyx import dumps, loads
from jsonyx.scanner import JSONSyntaxError, format_error
from typing_extensions import Any


class JSONNamespace:  # pylint: disable=R0903
    """JSON namespace."""

    allow: list[str]
    compact: bool
    ensure_ascii: bool
    indent: int | str | None
    filename: str | None
    sort_keys: bool


def register(parser: ArgumentParser) -> None:
    """Register JSON tool."""
    parser.add_argument("filename", nargs="?")
    parser.add_argument("--compact", action="store_true")
    parser.add_argument("--ensure-ascii", action="store_true")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--indent", type=int, metavar="SPACES")
    group.add_argument(
        "--tab",
        action="store_const",
        const="\t",
        dest="indent",
        help="indent using tabs",
    )
    parser.add_argument(
        "--nonstrict",
        action="store_const",
        const=["comments", "duplicate_keys", "nan", "trailing_comma"],
        default=[],
        dest="allow",
    )
    parser.add_argument("--sort-keys", action="store_true")


def run(args: JSONNamespace) -> None:
    """Run JSON tool."""
    input_json: bytes | str
    if args.filename:
        filename: str = args.filename
        input_json = Path(filename).read_bytes()
    elif stdin.isatty():
        filename, input_json = "<stdin>", "\n".join(iter(input, ""))
    else:
        filename, input_json = "<stdin>", stdin.buffer.read()

    try:
        obj: Any = loads(input_json, allow=args.allow, filename=filename)
    except JSONSyntaxError as exc:
        raise SystemExit(format_error(exc)) from None

    output_json: str = dumps(
        obj,
        allow=["nan"],
        ensure_ascii=args.ensure_ascii,
        indent=args.indent,
        item_separator="," if args.compact else ", ",
        key_separator=":" if args.compact else ": ",
        sort_keys=args.sort_keys,
    )
    print(output_json)


def _main() -> None:
    parser: ArgumentParser = ArgumentParser()
    register(parser)
    try:
        run(parser.parse_args(namespace=JSONNamespace()))
    except BrokenPipeError as exc:
        sys.exit(exc.errno)


if __name__ == "__main__":
    _main()



# Copyright (C) 2024 Nice Zombies
"""JSON accelerator."""
__all__: list[str] = ["DuplicateKey", "make_scanner", "scanstring"]

from collections.abc import Callable

from jsonyx.decoder import JSONDecoder
from typing_extensions import Any


def make_scanner(context: JSONDecoder) -> (
    Callable[[str, str, int], tuple[Any, int]]
):
    """Make JSON scanner."""


def scanstring(filename: str, s: str, end: int, /) -> tuple[str, int]:
    """Scan JSON string."""


class DuplicateKey(str):  # noqa: SLOT000
    """Duplicate key."""



# Copyright (C) 2024 Nice Zombies
"""JSON scanner."""
from __future__ import annotations

__all__ = ["JSONSyntaxError", "format_error", "make_scanner"]

import re
from math import inf, nan
from re import DOTALL, MULTILINE, VERBOSE, Match
from shutil import get_terminal_size
from typing import TYPE_CHECKING

from typing_extensions import Any

if TYPE_CHECKING:
    from collections.abc import Callable

    from jsonyx.decoder import JSONDecoder

_match_number: Callable[[str, int], Match[str] | None] = re.compile(
    r"(-?(?:0|[1-9]\d*))(\.\d+)?([eE][-+]?\d+)?", VERBOSE | MULTILINE | DOTALL,
).match


def _get_err_context(doc: str, pos: int) -> tuple[int, str]:
    line_start: int = doc.rfind("\n", 0, pos) + 1
    if (line_end := doc.find("\n", pos)) == -1:
        line_end = len(doc)

    max_chars: int = get_terminal_size().columns - 4  # leading spaces
    min_start: int = min(line_end - max_chars, pos - max_chars // 2)
    max_end: int = max(line_start + max_chars, pos + max_chars // 2)
    text_start: int = max(min_start, line_start)
    text_end: int = min(line_end, max_end)
    text: str = doc[text_start:text_end].expandtabs(1)
    if text_start > line_start:
        text = "..." + text[3:]

    if text_end < line_end:
        text = text[:-3] + "..."

    offset: int = pos - text_start + 1
    return offset, text


def format_error(exc: JSONSyntaxError) -> str:
    """Format JSON syntax error."""
    caret_line: str = " " * (exc.offset - 1) + "^"  # type: ignore
    exc_type: type[JSONSyntaxError] = type(exc)
    return f"""\
  File {exc.filename!r}, line {exc.lineno:d}, column {exc.colno:d}
    {exc.text}
    {caret_line}
{exc_type.__module__}.{exc_type.__qualname__}: {exc.msg}\
"""


class JSONSyntaxError(SyntaxError):
    """JSON syntax error."""

    def __init__(self, msg: str, filename: str, doc: str, pos: int) -> None:
        """Create new JSON syntax error."""
        lineno: int = doc.count("\n", 0, pos) + 1
        self.colno: int = pos - doc.rfind("\n", 0, pos)
        offset, text = _get_err_context(doc, pos)
        super().__init__(msg, (filename, lineno, offset, text))

    def __str__(self) -> str:
        """Convert to string."""
        return (
            f"{self.msg} (file {self.filename}, line {self.lineno:d}, column "
            f"{self.colno: d})"
        )


try:
    from jsonyx._accelerator import make_scanner
except ImportError:
    # pylint: disable-next=R0915
    def make_scanner(context: JSONDecoder) -> (   # noqa: C901, PLR0915
        Callable[[str, str, int], tuple[Any, int]]
    ):
        """Make JSON scanner."""
        allow_nan: bool = context.allow_nan
        memo: dict[str, str] = {}
        memoize: Callable[[str, str], str] = memo.setdefault
        parse_array: Callable[[
            JSONDecoder, str, str, int,
            Callable[[str, str, int], tuple[Any, int]],
        ], tuple[list[Any], int]] = context.parse_array
        parse_object: Callable[[
            JSONDecoder, str, str, int,
            Callable[[str, str, int], tuple[Any, int]],
            Callable[[str, str], str],
        ], tuple[dict[str, Any], int]] = context.parse_object
        parse_string: Callable[
            [str, str, int], tuple[str, int],
        ] = context.parse_string

        # pylint: disable-next=R0912
        def _scan_once(  # noqa: C901, PLR0912
            filename: str, s: str, idx: int,
        ) -> tuple[Any, int]:
            try:
                nextchar = s[idx]
            except IndexError:
                msg: str = "Expecting value"
                raise JSONSyntaxError(msg, filename, s, idx) from None

            result: Any
            if nextchar == '"':
                result, end = parse_string(filename, s, idx + 1)
            elif nextchar == "{":
                result, end = parse_object(
                    context, filename, s, idx + 1, _scan_once, memoize,
                )
            elif nextchar == "[":
                result, end = parse_array(
                    context, filename, s, idx + 1, _scan_once,
                )
            elif nextchar == "n" and s[idx:idx + 4] == "null":
                result, end = None, idx + 4
            elif nextchar == "t" and s[idx:idx + 4] == "true":
                result, end = True, idx + 4
            elif nextchar == "f" and s[idx:idx + 5] == "false":
                result, end = False, idx + 5
            elif number := _match_number(s, idx):
                integer, frac, exp = number.groups()
                end = number.end()
                if frac or exp:
                    result = float(integer + (frac or "") + (exp or ""))
                else:
                    result = int(integer)
            elif nextchar == "N" and s[idx:idx + 3] == "NaN":
                if not allow_nan:
                    msg: str = "NaN isn't allowed"
                    raise JSONSyntaxError(msg, filename, s, idx)

                result, end = nan, idx + 3
            elif nextchar == "I" and s[idx:idx + 8] == "Infinity":
                if not allow_nan:
                    msg = "Infinity isn't allowed"
                    raise JSONSyntaxError(msg, filename, s, idx)

                result, end = inf, idx + 8
            elif nextchar == "-" and s[idx:idx + 9] == "-Infinity":
                if not allow_nan:
                    msg = "Infinity isn't allowed"
                    raise JSONSyntaxError(msg, filename, s, idx)

                result, end = -inf, idx + 9
            else:
                msg = "Expecting value"
                raise JSONSyntaxError(msg, filename, s, idx)

            return result, end

        def scan_once(filename: str, s: str, idx: int) -> tuple[Any, int]:
            try:
                return _scan_once(filename, s, idx)
            finally:
                memo.clear()

        return scan_once
